{$IF DEFINED(PATCH_HEART) and DEFINED(SDL2)}
unit qrl_viewer;
{ Quasi Robot Language }

interface

uses sysutils,
	SDL2, SDL2_Image, SDL2_TTF;


Function QRL_Viewer_EventDriven( event: TSDL_Event ): Boolean;

Function QRL_Viewer_InitTracer( name: String ): Boolean;
Function QRL_Viewer_StartTracer: Boolean;
Function QRL_Viewer_ShowTracer( msg: String ): Boolean;
Function QRL_Viewer_RefreshTracer: Boolean;
Function QRL_Viewer_StopTracer: Boolean;



implementation

uses
	math,
	errmsg,
{$IFDEF PATCH_GH}
	gears_base,
{$ENDIF PATCH_GH}
	gears,
{$IFDEF SDLMODE}
	sdlinit,
	sdlgfx,
{$ELSE SDLMODE}
	ascii_sdlinit,
	ascii_sdl,
{$ENDIF SDLMODE}
	texutil,
	qrl_init,
	ui4gh;

type
	QRL_Tracer = Record
		X: Integer;
		Y: Integer;
	end;

const
	QRLViewer_WindowName: PChar = 'GearHead Arena QRL Viewer';

	QRLViewer_FontName = Graphics_Directory + 'Vera.ttf';
	QRLViewer_FontSize = 11;

	QRLViewer_Graphics_DirName = 'Image.qrlmaker';
	QRLViewer_Graphics_Directory = QRLViewer_Graphics_Dirname + OS_Dir_Separator;
	QRLViewer_Graphics_Cursor_Focus = QRLViewer_Graphics_Directory + 'cursor_focus.png';
	QRLViewer_Graphics_Cursor_Select = QRLViewer_Graphics_Directory + 'cursor_select.png';
	QRLViewer_Graphics_Cursor_Highlight = QRLViewer_Graphics_Directory + 'cursor_highlight.png';

{$IF DEFINED(CPU32)}
	QRLViewer_GaugeSize = 16;
	QRLViewer_GapSize = 8;
	QRLViewer_CoreSize = 32;
{$ELSE}
	QRLViewer_GaugeSize = 32;
	QRLViewer_GapSize = 16;
	QRLViewer_CoreSize = 64;
{$ENDIF}
	QRLViewer_Count_X = 100;
	QRLViewer_Count_Y = 100;
{$IF DEFINED(CPU32)}
	QRLViewer_Scaling_dB_default = 0;
{$ELSE}
	QRLViewer_Scaling_dB_default = -5;
{$ENDIF}
	QRLViewer_Scaling_dB_Min = -40;
	QRLViewer_Scaling_dB_Max =  20;

var
	QRLViewer_Window: PSDL_Window;
	QRLViewer_Renderer: PSDL_Renderer;

	QRLViewer_Font: PTTF_Font;

	QRLViewer_Image_Cursor_Focus: PSDL_Texture;
	QRLViewer_Image_Cursor_Select: PSDL_Texture;
	QRLViewer_Image_Cursor_Highlight: PSDL_Texture;

	QRLViewer_ImageFileName: String;
	QRLViewer_Image: PSDL_Texture;

	QRLViewer_Image_Offset_X: Integer;
	QRLViewer_Image_Offset_Y: Integer;
	QRLViewer_Image_Scaling_dB: Integer;
	QRLViewer_Image_Scaling: Single;

	QRLViewer_Tracer: Array [1..QRL_ChipMode_LoopCount_MAX , 1..QRL_ChipMode_NumOfSquares_MAX] of QRL_Tracer;
	QRLViewer_Tracer_Num: Integer;
	QRLViewer_Tracer_LastNum: Integer;


Function QRLViewer_CreateTextureFromIMG( name: String ): PSDL_Texture;
var
	tmp_name: PChar;
	tmp_Surface: PSDL_Surface;
{$IF DEFINED(CPU32)}
	tmp_Surface2: PSDL_Surface;
	ret: Integer;
{$ENDIF}
	tmp_Texture: PSDL_Texture;
{$IF DEFINED(SDL2) and (DEFINED(FreeBSD) or DEFINED(LINUX))}
	FPUExceptionMask: TFPUExceptionMask;
{$ENDIF}
begin
	tmp_name := QuickPCopy( name );
	tmp_Surface := IMG_Load( tmp_name );
	sysutils.StrDispose( tmp_name );
	if ( NIL = tmp_Surface ) then begin
		ErrorMessage('IMG_Load(' + name + ') failed.');
		exit(NIL);
	end;

{$IF DEFINED(CPU32)}
	tmp_Surface2 := SDL_CreateRGBSurface( 0 , (tmp_Surface^.W div 2) , (tmp_Surface^.H div 2) , tmp_Surface^.Format^.BitsPerPixel , tmp_Surface^.Format^.Rmask , tmp_Surface^.Format^.Gmask , tmp_Surface^.Format^.Bmask , tmp_Surface^.Format^.Amask );
	if ( NIL = tmp_Surface2 ) then begin
		ErrorMessage('SDL_CreateRGBSurface() failed. : ' + SDL_GetError);
		SDL_FreeSurface( tmp_Surface );
		exit(NIL);
	end;
	ret := SDL_BlitSurfaceScaled( tmp_Surface , NIL , tmp_Surface2 , NIL );
	if ( ret < 0 ) then begin
		ErrorMessage('SDL_BlitSurfaceScaled() failed. : ' + SDL_GetError);
		SDL_FreeSurface( tmp_Surface2 );
		SDL_FreeSurface( tmp_Surface );
		exit(NIL);
	end;

	SDL_FreeSurface( tmp_Surface );
	tmp_Surface := tmp_Surface2;
{$ENDIF}

	if ( 0 <> tmp_Surface^.format^.Amask ) then begin
{$IF DEFINED(SDL2) and (DEFINED(FreeBSD) or DEFINED(LINUX))}
		FPUExceptionMask := GetExceptionMask;
		SetExceptionMask( FPUExceptionMask + [exInvalidOp, exZeroDivide] );
{$ENDIF}
		QRLViewer_CreateTextureFromIMG := SDL_CreateTextureFromSurface( QRLViewer_Renderer, tmp_Surface );
{$IF DEFINED(SDL2) and (DEFINED(FreeBSD) or DEFINED(LINUX))}
		SetExceptionMask( FPUExceptionMask );
{$ENDIF}
		if ( NIL = QRLViewer_CreateTextureFromIMG ) then begin
			ErrorMessage('SDL_CreateTextureFromSurface() with Amask failed. : ' + SDL_GetError);
			SDL_FreeSurface( tmp_Surface );
			exit(NIL);
		end;
	end else begin
		tmp_Texture := SDL_CreateTextureFromSurface( QRLViewer_Renderer , tmp_Surface );
		if ( NIL = tmp_Texture ) then begin
			ErrorMessage('SDL_CreateTextureFromSurface() failed. : ' + SDL_GetError);
			SDL_FreeSurface( tmp_Surface );
			exit(NIL);
		end;
		QRLViewer_CreateTextureFromIMG := SDL_CreateTexture( QRLViewer_Renderer , SDL_PIXELFORMAT_RGB24 , SDL_TEXTUREACCESS_TARGET , tmp_Surface^.W , tmp_Surface^.H );
		if ( NIL = QRLViewer_CreateTextureFromIMG ) then begin
			ErrorMessage('SDL_CreateTexture() failed. : ' + SDL_GetError);
			SDL_DestroyTexture( tmp_Texture );
			SDL_FreeSurface( tmp_Surface );
			exit(NIL);
		end;
		SDL_SetRenderTarget( QRLViewer_Renderer , QRLViewer_CreateTextureFromIMG );
		SDL_RenderCopy( QRLViewer_Renderer , tmp_Texture , NIL , NIL );
		SDL_DestroyTexture( tmp_Texture );
		SDL_SetRenderTarget( QRLViewer_Renderer , NIL );
	end;

	SDL_FreeSurface( tmp_Surface );
end;

Function RenderText( const pline: String; var fg: TSDL_Color ): PSDL_Texture;
var
	pWC: PUInt16;
	bg: TSDL_Color;
	tmp_Surface: PSDL_Surface;
begin
	pWC := New_Conv_ToUni16( pline );
	if not(SDL_AAFont) and not(SDL_AAFont_Shaded) then begin
		tmp_Surface := TTF_RenderUnicode_Solid( QRLViewer_Font, pWC, fg );
		RenderText := SDL_CreateTextureFromSurface( QRLViewer_Renderer , tmp_Surface );
	end else if SDL_AAFont_Shaded then begin
		bg.r := 0;
		bg.b := 0;
		bg.g := 0;
		tmp_Surface := TTF_RenderUnicode_Shaded( QRLViewer_Font, pWC, fg, bg );
		RenderText := SDL_CreateTextureFromSurface( QRLViewer_Renderer , tmp_Surface );
	end else begin
		tmp_Surface := TTF_RenderUnicode_Blended( QRLViewer_Font, pWC, fg );
		RenderText := SDL_CreateTextureFromSurface( QRLViewer_Renderer , tmp_Surface );
	end;
	SDL_FreeSurface( tmp_Surface );
	sysutils.StrDispose( PChar( pWC ) );
end;

Procedure QRL_Viewer_Flip;
var
	DrawingWidth, DrawingHeight: Integer;
	iw, ih: Integer;
	MySrc, MyDest: TSDL_Rect;
	i: Integer;
	LimitPosition: Integer;
	MyImage: PSDL_Texture;
{$IF DEFINED(SDL2) and (DEFINED(FreeBSD) or DEFINED(LINUX))}
	FPUExceptionMask: TFPUExceptionMask;
{$ENDIF}
begin
	if ( QRLViewer_Image_Offset_X < 1 ) or ( QRLViewer_Image_Offset_Y < 1 ) or ( QRLViewer_Count_X < QRLViewer_Image_Offset_X ) or ( QRLViewer_Count_Y < QRLViewer_Image_Offset_Y ) then begin
		exit;
	end;
	if ( QRLViewer_Image_Scaling_dB < QRLViewer_Scaling_dB_Min ) or ( QRLViewer_Scaling_dB_Max < QRLViewer_Image_Scaling_dB ) then begin
		exit;
	end;
	QRLViewer_Image_Scaling := Power(10.0, (Single(QRLViewer_Image_Scaling_dB) / 20.0));

	DrawingWidth := QRLViewer_ScreenSize_Width - QRLViewer_GaugeSize - SDL_Scrollbar_Size;
	DrawingHeight := QRLViewer_ScreenSize_Height - QRLViewer_GaugeSize - SDL_Scrollbar_Size;

	{ Show Main Graphics }
	MyDest.X := QRLViewer_GaugeSize;
	MyDest.Y := QRLViewer_GaugeSize;
	MyDest.W := DrawingWidth;
	MyDest.H := DrawingHeight;

	MySrc.X := QRLViewer_GaugeSize + ( QRLViewer_Image_Offset_X - 1 ) * ( QRLViewer_GapSize + QRLViewer_CoreSize );
	MySrc.Y := QRLViewer_GaugeSize + ( QRLViewer_Image_Offset_Y - 1 ) * ( QRLViewer_GapSize + QRLViewer_CoreSize );
	MySrc.W := Round(Single(MyDest.W) / QRLViewer_Image_Scaling);
	MySrc.H := Round(Single(MyDest.H) / QRLViewer_Image_Scaling);

	SDL_SetRenderDrawColor( QRLViewer_Renderer , 0 , 0 , 0 , SDL_ALPHA_OPAQUE );
	SDL_RenderFillRect( QRLViewer_Renderer , @MyDest );
	SDL_QueryTexture( QRLViewer_Image , NIL , NIL , @iw , @ih );
	if ( iw < ( MySrc.X + MySrc.W ) ) then begin
		MyDest.W := Round(Single( iw - MySrc.X ) * QRLViewer_Image_Scaling);
	end;
	if ( ih < ( MySrc.Y + MySrc.H ) ) then begin
		MyDest.H := Round(Single( ih - MySrc.Y ) * QRLViewer_Image_Scaling);
	end;
	SDL_RenderCopy( QRLViewer_Renderer , QRLViewer_Image , @MySrc , @MyDest );

	{ Show Gauge X }
	MyDest.X := QRLViewer_GaugeSize;
	MyDest.Y := 0;
	MyDest.W := DrawingWidth;
	MyDest.H := QRLViewer_GaugeSize;
	SDL_SetRenderDrawColor( QRLViewer_Renderer , 0 , 0 , 0 , SDL_ALPHA_OPAQUE );
	SDL_RenderFillRect( QRLViewer_Renderer , @MyDest );
	LimitPosition := QRLViewer_ScreenSize_Width - SDL_Scrollbar_Size;
	i := QRLViewer_Image_Offset_X;
	repeat
		MyImage := RenderText( BStr(i) , StdWhite );
		SDL_QueryTexture( MyImage , NIL , NIL , @MyDest.W , @MyDest.H );
		MyDest.X := QRLViewer_GaugeSize + Round( Single( ( i - QRLViewer_Image_Offset_X ) * ( QRLViewer_GapSize + QRLViewer_CoreSize ) + QRLViewer_GapSize + ( ( QRLViewer_CoreSize - MyDest.W ) div 2 ) ) * QRLViewer_Image_Scaling );
		MyDest.Y := 0;
		if ( LimitPosition < ( MyDest.X + MyDest.W ) ) then break;
		SDL_RenderCopy( QRLViewer_Renderer , MyImage , NIL , @MyDest );
		SDL_DestroyTexture( MyImage );
		if (QRLViewer_Image_Scaling_dB < 0) then begin
			i := i + Round(1.0 / QRLViewer_Image_Scaling);
		end else begin
			inc( i );
		end;
	until ( QRLViewer_Count_X < i );

	{ Show Gauge Y }
	MyDest.X := 0;
	MyDest.Y := QRLViewer_GaugeSize;
	MyDest.W := QRLViewer_GaugeSize;
	MyDest.H := DrawingHeight;
	SDL_SetRenderDrawColor( QRLViewer_Renderer , 0 , 0 , 0 , SDL_ALPHA_OPAQUE );
	SDL_RenderFillRect( QRLViewer_Renderer , @MyDest );
	LimitPosition := QRLViewer_ScreenSize_Height - SDL_Scrollbar_Size;
	i := QRLViewer_Image_Offset_Y;
	repeat
		MyImage := RenderText( BStr(i) , StdWhite );
		SDL_QueryTexture( MyImage , NIL , NIL , @MyDest.W , @MyDest.H );
		MyDest.X := 0;
		MyDest.Y := QRLViewer_GaugeSize + Round( Single( ( i - QRLViewer_Image_Offset_Y ) * ( QRLViewer_GapSize + QRLViewer_CoreSize ) + QRLViewer_GapSize + ( ( QRLViewer_CoreSize - MyDest.H ) div 2 ) ) * QRLViewer_Image_Scaling );
		if ( LimitPosition < ( MyDest.Y + MyDest.H ) ) then break;
		SDL_RenderCopy( QRLViewer_Renderer , MyImage , NIL , @MyDest );
		SDL_DestroyTexture( MyImage );
		if (QRLViewer_Image_Scaling_dB < 0) then begin
			i := i + Round(1.0 / QRLViewer_Image_Scaling);
		end else begin
			inc( i );
		end;
	until ( QRLViewer_Count_Y < i );

	{ Show Scroll Bar X }
	MyDest.X := QRLViewer_GaugeSize;
	MyDest.Y := QRLViewer_ScreenSize_Height - SDL_Scrollbar_Size;
	MyDest.W := DrawingWidth;
	MyDest.H := SDL_Scrollbar_Size;
	SDL_SetRenderDrawColor( QRLViewer_Renderer , MenuItem.R , MenuItem.G , MenuItem.B , SDL_ALPHA_OPAQUE );
	SDL_RenderFillRect( QRLViewer_Renderer , @MyDest );
	MyDest.X := QRLViewer_GaugeSize + DrawingWidth * ( QRLViewer_Image_Offset_X - 1 ) div QRLViewer_Count_X;
	MyDest.W := DrawingWidth * Round( Single(DrawingWidth) / QRLViewer_Image_Scaling ) div ( ( QRLViewer_GapSize + QRLViewer_CoreSize ) * QRLViewer_Count_X + QRLViewer_GapSize );
	if ( ( QRLViewer_ScreenSize_Width - SDL_Scrollbar_Size ) < ( MyDest.X + MyDest.W ) ) then begin
		MyDest.W := ( QRLViewer_ScreenSize_Width - SDL_Scrollbar_Size ) - MyDest.X;
	end;
	SDL_SetRenderDrawColor( QRLViewer_Renderer , MenuSelect.R , MenuSelect.G , MenuSelect.B , SDL_ALPHA_OPAQUE );
	SDL_RenderFillRect( QRLViewer_Renderer , @MyDest );

	{ Show Scroll Bar Y }
	MyDest.X := QRLViewer_ScreenSize_Width - SDL_Scrollbar_Size;
	MyDest.Y := QRLViewer_GaugeSize;
	MyDest.W := SDL_Scrollbar_Size;
	MyDest.H := DrawingHeight;
	SDL_SetRenderDrawColor( QRLViewer_Renderer , MenuItem.R , MenuItem.G , MenuItem.B , SDL_ALPHA_OPAQUE );
	SDL_RenderFillRect( QRLViewer_Renderer , @MyDest );
	MyDest.Y := QRLViewer_GaugeSize + DrawingHeight * ( QRLViewer_Image_Offset_Y - 1 ) div QRLViewer_Count_Y;
	MyDest.H := DrawingHeight * Round( Single(DrawingHeight) / QRLViewer_Image_Scaling ) div ( ( QRLViewer_GapSize + QRLViewer_CoreSize ) * QRLViewer_Count_Y + QRLViewer_GapSize );
	if ( ( QRLViewer_ScreenSize_Height - SDL_Scrollbar_Size ) < ( MyDest.Y + MyDest.H ) ) then begin
		MyDest.H := ( QRLViewer_ScreenSize_Height - SDL_Scrollbar_Size ) - MyDest.Y;
	end;
	SDL_SetRenderDrawColor( QRLViewer_Renderer , MenuSelect.R , MenuSelect.G , MenuSelect.B , SDL_ALPHA_OPAQUE );
	SDL_RenderFillRect( QRLViewer_Renderer , @MyDest );

	{ Finish }
{$IF DEFINED(SDL2) and (DEFINED(FreeBSD) or DEFINED(LINUX))}
	FPUExceptionMask := GetExceptionMask;
	SetExceptionMask( FPUExceptionMask + [exZeroDivide] );
{$ENDIF}
	SDL_RenderPresent( QRLViewer_Renderer );
{$IF DEFINED(SDL2) and (DEFINED(FreeBSD) or DEFINED(LINUX))}
	SetExceptionMask( FPUExceptionMask );
{$ENDIF}
end;

Function QRL_Viewer_EventDriven( event: TSDL_Event ): Boolean;
	Function ScrollBar4Mouse( X, Y: Integer ): Boolean;
	var
		flip: Boolean;
	begin
		flip := False;
		if ( ( ( QRLViewer_ScreenSize_Height - SDL_Scrollbar_Size ) <= Y ) and ( QRLViewer_GaugeSize <= X ) ) then begin
			{ Handle Scroll Bars X }
			QRLViewer_Image_Offset_X := ( X - QRLViewer_GaugeSize ) * QRLViewer_Count_X div ( QRLViewer_ScreenSize_Width - QRLViewer_GaugeSize - SDL_Scrollbar_Size );
			if ( QRLViewer_Image_Offset_X < 1 ) then begin
				QRLViewer_Image_Offset_X := 1;
			end else if ( QRLViewer_Count_X < QRLViewer_Image_Offset_X ) then begin
				QRLViewer_Image_Offset_X := QRLViewer_Count_X;
			end;
			flip := True;
		end;
		if ( ( ( QRLViewer_ScreenSize_Width - SDL_Scrollbar_Size ) <= X ) and ( QRLViewer_GaugeSize <= Y ) ) then begin
			{ Handle Scroll Bars Y }
			QRLViewer_Image_Offset_Y := ( Y - QRLViewer_GaugeSize ) * QRLViewer_Count_Y div ( QRLViewer_ScreenSize_Height - QRLViewer_GaugeSize - SDL_Scrollbar_Size );
			if ( QRLViewer_Image_Offset_Y < 1 ) then begin
				QRLViewer_Image_Offset_Y := 1;
			end else if ( QRLViewer_Count_Y < QRLViewer_Image_Offset_Y ) then begin
				QRLViewer_Image_Offset_Y := QRLViewer_Count_Y;
			end;
			flip := True;
		end;
		ScrollBar4Mouse := flip;
	end;
var
	flip: Boolean;
	x_m, x_p, y_m, y_p, m_m, m_p, m_r: Boolean;
	X, Y, top, left, bottom, right: Integer;
	width, height: Integer;
begin
	flip := False;
	case event.type_ of

	SDL_KeyDown:
		begin
			x_m := False; x_p := False;
			y_m := False; y_p := False;
			m_m := False; m_p := False; m_r := False;

			case event.key.keysym.sym of
			SDLK_Up, SDLK_KP_8, SDLK_k:	begin y_m := True; end;
			SDLK_Down, SDLK_KP_2, SDLK_j:	begin y_p := True; end;
			SDLK_Left, SDLK_KP_4, SDLK_h:	begin x_m := True; end;
			SDLK_Right, SDLK_KP_6, SDLK_l:	begin x_p := True; end;
			SDLK_KP_7, SDLK_y:		begin x_m := True; y_m := True; end;
			SDLK_KP_9, SDLK_u:		begin x_p := True; y_m := True; end;
			SDLK_KP_1, SDLK_b:		begin x_m := True; y_p := True; end;
			SDLK_KP_3, SDLK_n:		begin x_p := True; y_p := True; end;
			SDLK_PageUp:			begin m_m := True; end;
			SDLK_PageDown:			begin m_p := True; end;
			SDLK_0:				begin m_r := True; end;
			SDLK_KP_0:			begin m_r := True; end;
			end;

			if ( x_m and ( 1 < QRLViewer_Image_Offset_X ) ) then begin
				dec( QRLViewer_Image_Offset_X );
				flip := True;
			end;
			if ( x_p and ( QRLViewer_Image_Offset_X < QRLViewer_Count_X ) ) then begin
				inc( QRLViewer_Image_Offset_X );
				flip := True;
			end;
			if ( y_m and ( 1 < QRLViewer_Image_Offset_Y ) ) then begin
				dec( QRLViewer_Image_Offset_Y );
				flip := True;
			end;
			if ( y_p and ( QRLViewer_Image_Offset_Y < QRLViewer_Count_Y ) ) then begin
				inc( QRLViewer_Image_Offset_Y );
				flip := True;
			end;
			if ( m_m and ( QRLViewer_Image_Scaling_dB < QRLViewer_Scaling_dB_Max ) ) then begin
				inc( QRLViewer_Image_Scaling_dB );
				flip := True;
			end;
			if ( m_p and ( QRLViewer_Scaling_dB_Min < QRLViewer_Image_Scaling_dB ) ) then begin
				dec( QRLViewer_Image_Scaling_dB );
				flip := True;
			end;
			if ( m_r ) then begin
				QRLViewer_Image_Scaling_dB := QRLViewer_Scaling_dB_default;
				flip := True;
			end;
		end;

	SDL_MouseWheel:
		begin
			m_m := False; m_p := False;

			if ( 0 < event.wheel.x ) then begin
				m_m := True;
			end else if ( event.wheel.x < 0 ) then begin
				m_p := True;
			end;
			if ( 0 < event.wheel.y ) then begin
				m_m := True;
			end else if ( event.wheel.y < 0 ) then begin
				m_p := True;
			end;

			if ( m_m and ( QRLViewer_Image_Scaling_dB < QRLViewer_Scaling_dB_Max ) ) then begin
				inc( QRLViewer_Image_Scaling_dB );
				flip := True;
			end;
			if ( m_p and ( QRLViewer_Scaling_dB_Min < QRLViewer_Image_Scaling_dB ) ) then begin
				dec( QRLViewer_Image_Scaling_dB );
				flip := True;
			end;
		end;

	SDL_MouseButtonDown:
		begin
			case event.button.button of
			SDL_Button_Left:
				begin
					if ScrollBar4Mouse( event.button.x , event.button.y ) then begin
						flip := True;
					end;
				end;

			SDL_Button_Middle:	;
			SDL_Button_Right:	;
			end;
		end;

	SDL_MouseMotion:
		begin
			if ( 0 <> ( SDL_BUTTON_LMASK and event.motion.state ) ) then begin
				if ScrollBar4Mouse( event.motion.x , event.motion.y ) then begin
					flip := True;
				end;
			end;
		end;

	SDL_WindowEvent:
		begin
			case event.window.event of
			SDL_WindowEvent_Exposed:
				begin
					flip := True;
				end;

			SDL_WindowEvent_Moved:
				begin
					if ( -1 <> QRLViewer_ScreenPos_X ) and ( -1 <> QRLViewer_ScreenPos_Y ) then begin
						SDL_GetWindowBordersSize( QRLViewer_Window , @top , @left , @bottom , @right );
						QRLViewer_ScreenPos_X := event.window.data1 - left;
						QRLViewer_ScreenPos_Y := event.window.data2 - top;
						flip := True;
					end;
				end;

			SDL_WindowEvent_Resized:
				begin
					width := event.window.data1;
					if width < 320 then width := 320;
					height := event.window.data2;
					if height < 240 then height := 240;

					QRLViewer_ScreenSize_Width  := width;
					QRLViewer_ScreenSize_Height := height;

					SDL_SetWindowSize( QRLViewer_Window , width , height );

					SDL_GetWindowPosition( QRLViewer_Window , @X , @Y );
					SDL_GetWindowBordersSize( QRLViewer_Window , @top , @left , @bottom , @right );
					QRLViewer_ScreenPos_X := X - left;
					QRLViewer_ScreenPos_Y := Y - top;

					flip := True;
				end;
			end;
		end;
	end;

	if flip then begin
		QRL_Viewer_Flip;
	end;

	QRL_Viewer_EventDriven := True;
end;

Procedure QRL_Viewer_ShowCursor( cursor: PSDL_Texture; X, Y: Integer );
var
	MyDest: TSDL_Rect;
begin
	MyDest.X := QRLViewer_GaugeSize + ( X - 1 ) * ( QRLViewer_GapSize + QRLViewer_CoreSize );
	MyDest.Y := QRLViewer_GaugeSize + ( Y - 1 ) * ( QRLViewer_GapSize + QRLViewer_CoreSize );
	SDL_QueryTexture( cursor , NIL , NIL , @MyDest.W , @MyDest.H );

	SDL_SetRenderTarget( QRLViewer_Renderer , QRLViewer_Image );
	SDL_RenderCopy( QRLViewer_Renderer , cursor , NIL , @MyDest );
	SDL_SetRenderTarget( QRLViewer_Renderer , NIL );
end;

Function QRL_Viewer_InitTracer( name: String ): Boolean;
var
	userfilename: String;
begin
	userfilename := QRL_UserDirectory + name + QRL_ImageFileExtName;
	if FileExists( userfilename ) then begin
		QRLViewer_ImageFileName := userfilename;
	end else begin
		QRLViewer_ImageFileName := QRL_SystemDirectory + name + QRL_ImageFileExtName;
	end;

	QRLViewer_Tracer_Num := 0;
	QRLViewer_Tracer_LastNum := 0;

	if ( NIL <> QRLViewer_Image ) then begin
		SDL_DestroyTexture( QRLViewer_Image );
	end;
	QRLViewer_Image := QRLViewer_CreateTextureFromIMG( QRLViewer_ImageFileName );

	QRL_Viewer_Flip;

	QRL_Viewer_InitTracer := True;
end;

Function QRL_Viewer_StartTracer: Boolean;
begin
	QRLViewer_Tracer_Num := 0;
	QRLViewer_Tracer_LastNum := 0;

	QRL_Viewer_StartTracer := True;
end;

Function QRL_Viewer_ShowTracer( msg: String ): Boolean;
var
	i, j: Integer;
	P_hyphen, P_comma: SizeInt;
	X, Y: Integer;
begin
	for i := 1 to QRLViewer_Tracer_LastNum do begin
		for j := 1 to QRL_ChipMode_NumOfSquares_MAX do begin
			if ( ( QRLViewer_Tracer[i,j].X < 1 ) or ( QRLViewer_Tracer[i,j].Y < 1 ) ) then break;
			QRL_Viewer_ShowCursor( QRLViewer_Image_Cursor_Highlight , QRLViewer_Tracer[i,j].X , QRLViewer_Tracer[i,j].Y );
		end;
	end;
	QRLViewer_Tracer_LastNum := 0;

	if ( Length(msg) < 1 ) then begin
		exit(True);
	end;
	if ( QRL_ChipMode_LoopCount_MAX <= QRLViewer_Tracer_Num ) then begin
		exit(False);
	end;

	if ( 0 < QRLViewer_Tracer_Num ) then begin
		for j := 1 to QRL_ChipMode_NumOfSquares_MAX do begin
			if ( ( QRLViewer_Tracer[ QRLViewer_Tracer_Num , j ].X < 1 ) or ( QRLViewer_Tracer[ QRLViewer_Tracer_Num , j ].Y < 1 ) ) then break;
			QRL_Viewer_ShowCursor( QRLViewer_Image_Cursor_Select , QRLViewer_Tracer[ QRLViewer_Tracer_Num , j ].X , QRLViewer_Tracer[ QRLViewer_Tracer_Num , j ].Y );
		end;
	end;

	inc( QRLViewer_Tracer_Num );

	j := 1;
	for j := 1 to QRL_ChipMode_NumOfSquares_MAX do begin
		P_hyphen := Pos('-', msg);
		P_comma := Pos(',', msg);
		if ( P_hyphen < 1 ) or ( P_comma < P_hyphen ) then begin
			QRLViewer_Tracer[ QRLViewer_Tracer_Num , j ].X := 0;
			QRLViewer_Tracer[ QRLViewer_Tracer_Num , j ].Y := 0;
			break;
		end;

		X := StrToInt( Copy( msg , 1 , ( P_hyphen - 1 ) ) );
		Y := StrToInt( Copy( msg , ( P_hyphen + 1 ) , ( P_comma - P_hyphen - 1 ) ) );

		QRLViewer_Tracer[ QRLViewer_Tracer_Num , j ].X := X;
		QRLViewer_Tracer[ QRLViewer_Tracer_Num , j ].Y := Y;

		QRL_Viewer_ShowCursor( QRLViewer_Image_Cursor_Focus , X , Y );

		msg := Copy( msg , ( P_comma + 1 ) , ( Length(msg) - P_comma ) );
	end;

	QRL_Viewer_Flip;

	QRL_Viewer_ShowTracer := True;
end;

Function QRL_Viewer_RefreshTracer: Boolean;
begin
	QRLViewer_Tracer_LastNum := QRLViewer_Tracer_Num;
	QRLViewer_Tracer_Num := 0;

	QRL_Viewer_RefreshTracer := True;
end;

Function QRL_Viewer_StopTracer: Boolean;
begin
	QRL_Viewer_ShowTracer( '' );
	QRLViewer_Tracer_Num := 0;

	QRL_Viewer_StopTracer := True;
end;


Procedure InitVideo;
{$IF DEFINED(SDL2) and (DEFINED(LINUX))}
var
	FPUExceptionMask: TFPUExceptionMask;
{$ENDIF}
begin
	if ( ( -1 <> QRLViewer_ScreenPos_X ) and ( -1 <> QRLViewer_ScreenPos_Y ) ) then begin
		QRLViewer_Window := SDL_CreateWindow(QRLViewer_WindowName, QRLViewer_ScreenPos_X, QRLViewer_ScreenPos_Y, QRLViewer_ScreenSize_Width, QRLViewer_ScreenSize_Height, ( SDL_WINDOW_RESIZABLE or SDL_WINDOW_SHOWN ) );
{$IFDEF WINDOWS}
		SDL_SetWindowPosition( QRLViewer_Window , QRLViewer_ScreenPos_X , QRLViewer_ScreenPos_Y );
{$ENDIF WINDOWS}
	end else begin
		QRLViewer_Window := SDL_CreateWindow(QRLViewer_WindowName, SDL_WINDOWPOS_UNDEFINED,SDL_WINDOWPOS_UNDEFINED, QRLViewer_ScreenSize_Width, QRLViewer_ScreenSize_Height, ( SDL_WINDOW_RESIZABLE or SDL_WINDOW_SHOWN ) );
	end;
	if (NIL = QRLViewer_Window) then begin
		ErrorMessage('SDL_CreateWindow() failed. : ' + SDL_GetError);
		halt(255);
	end;

{$IF DEFINED(SDL2) and (DEFINED(LINUX))}
	FPUExceptionMask := GetExceptionMask;
	SetExceptionMask( FPUExceptionMask + [exInvalidOp, exZeroDivide] );
{$ENDIF}
	QRLViewer_Renderer := SDL_CreateRenderer( QRLViewer_Window , -1 , SDL_RendererMode );
{$IF DEFINED(SDL2) and (DEFINED(LINUX))}
	SetExceptionMask( FPUExceptionMask );
{$ENDIF}
	if (NIL = QRLViewer_Renderer) then begin
		ErrorMessage('SDL_CreateRenderer() failed. : ' + SDL_GetError);
		halt(255);
	end;

	QRLViewer_WindowID := SDL_GetWindowID( QRLViewer_Window );

	QRLViewer_Font := TTF_OpenFont( QRLViewer_FontName , QRLViewer_FontSize );

	QRLViewer_Image_Cursor_Focus := QRLViewer_CreateTextureFromIMG( QRLViewer_Graphics_Cursor_Focus );
	QRLViewer_Image_Cursor_Select := QRLViewer_CreateTextureFromIMG( QRLViewer_Graphics_Cursor_Select );
	QRLViewer_Image_Cursor_Highlight := QRLViewer_CreateTextureFromIMG( QRLViewer_Graphics_Cursor_Highlight );

	QRLViewer_Image := NIL;
	QRLViewer_Image_Offset_X := 1;
	QRLViewer_Image_Offset_Y := 1;
	QRLViewer_Image_Scaling_dB := QRLViewer_Scaling_dB_default;
	QRLViewer_Image_Scaling := 0.0;
end;



initialization
begin
{$IFDEF DEBUG}
	ErrorMessage_fork('DEBUG: qrl_viewer.pp');
{$ENDIF DEBUG}

	InitVideo;
end;

finalization
begin
{$IFDEF DEBUG}
	ErrorMessage_fork('DEBUG: qrl_viewer.pp(finalization)');
{$ENDIF DEBUG}

	SDL_DestroyTexture( QRLViewer_Image_Cursor_Highlight );
	SDL_DestroyTexture( QRLViewer_Image_Cursor_Select );
	SDL_DestroyTexture( QRLViewer_Image_Cursor_Focus );

	TTF_CloseFont( QRLViewer_Font );

	if ( NIL <> QRLViewer_Image ) then begin
		SDL_DestroyTexture( QRLViewer_Image );
	end;

	SDL_DestroyRenderer( QRLViewer_Renderer );
end;

end.
{$ENDIF}
